#!/usr/bin/perl

##################################################################
#
#    -in_file=xxx      : xxx=input file name
#    -reg_map_file=xxx : xxx=output register map file name
#    -reg_cmd_file=xxx : xxx=output register command file name
#    -skip_byte=xxx    : xxx=skip byte for parser
#    -dmp_st=xxx       : xxx=dump start address (HEX)
#    -dmp_len=xxx      : xxx=dump length (HEX)
#
##################################################################
#
# For example :
# 
#
##################################################################


use IO::File;
use File::Path;

$dbg_verbose = 0;

$DUMP_START = 0x8000;
$DUMP_LEN   = 0x200;

  foreach $argv (@ARGV)
  {
        if($argv =~ s/-in_file=//)       { $IN_FILE       = $argv; }
     elsif($argv =~ s/-reg_map_file=//)  { $MAP_FILE      = $argv; }
     elsif($argv =~ s/-reg_cmd_file=//)  { $CMD_FILE      = $argv; }
     elsif($argv =~ s/-skip_byte=//)     { $skip_byte_num = $argv; }
     elsif($argv =~ s/-dmp_st=//)        { $DUMP_START    = hex($argv); }
     elsif($argv =~ s/-dmp_len=//)       { $DUMP_LEN      = hex($argv); }
     elsif($argv =~ /-help/)
     {
        $prog_name = $0;
        $prog_name =~ s/^.*\\//;
        $prog_name =~ s/^.*\///;
print <<EOF;
  -------------------------------------------
    $prog_name Help
  -------------------------------------------
    -in_file=xxx      : xxx=input file name
    -reg_map_file=xxx : xxx=output register map file name
    -reg_cmd_file=xxx : xxx=output register command file name
    -skip_byte=xxx    : xxx=skip byte for parser
    -dmp_st=xxx       : xxx=dump start address (HEX)
    -dmp_len=xxx      : xxx=dump length (HEX)
EOF

     exit();
     }
  }

  open(IN_BIN,$IN_FILE) or die "Cannot open file $IN_FILE due to $!";
  open(OUT_F,"> $MAP_FILE") or die("Can't open output file $OUT_FILE\n");
  open(CMD_F,"> $CMD_FILE") or die("Can't open output file $CMD_FILE\n");

  binmode(IN_BIN);

  $byte_cnt=0;


  $cur_bank = "NYET";

  $state = "NEW_CMD";

  while(read(IN_BIN,$byte,1))
  {

      if($byte_cnt>=$skip_byte_num)
      {
        $bin_raw = sprintf ("%2.2x",ord($byte));
        @nibble = split("",$bin_raw);

        if($dbg_verbose) { print CMD_F "$bin_raw --> "; }

        # state=new cmd
        if( $state eq "NEW_CMD" )
        {
          if( $nibble[1] eq "c" )
          {
            if($dbg_verbose) { print CMD_F "NEW_CMD  --> BANK=$nibble[0]\n"; }

            $cur_bank = $nibble[0];
          }
          elsif ($nibble[1] eq "a" ||
                 $nibble[1] eq "b" ||
                 $nibble[1] eq "c" ||
                 $nibble[1] eq "d" )
          {
             if($dbg_verbose) { print CMD_F "NEW_CMD  --> reserved\n"; }
          }
          elsif ($cur_bank eq "NYET")
          {
            die("Get Length info without BANK setting");
          }
          else
          {
            if($dbg_verbose) { print CMD_F "NEW_CMD  --> ADDR[11:8]=$nibble[0], LEN=$nibble[1]\n"; }

            $cur_len       = hex($nibble[1])+1;
            $cur_addr_11_8 = $nibble[0];

            $state="ADDR_7_0"

          }
        }
        elsif( $state eq "ADDR_7_0" )
        {
          $cur_addr_7_0 = $bin_raw;
          $cur_addr_11_0 = $cur_addr_7_0.$cur_addr_11_8;

          $addr_11_0_value = hex($cur_addr_11_0);

          if($dbg_verbose) { print CMD_F "ADDR_7_0 --> ADDR[7:0]=$cur_addr_7_0\n"; }
          
          $state="DATA";
          $cur_byte_cnt=0;

        }
        elsif( $state eq "DATA" )
        {
          $cur_addr_str = $cur_bank.sprintf("%3.3x", $addr_11_0_value);
          $data{$cur_addr_str} = uc($bin_raw);
          
          if($dbg_verbose) { print CMD_F "DATA     --> ADDR=$cur_addr_str, DATA=$data{$cur_addr_str}, byte_cnt=$cur_byte_cnt\n"; }
          else             { print CMD_F "ADDR=$cur_addr_str, DATA=$data{$cur_addr_str}\n"; }
          
          $cur_byte_cnt++;

          if($cur_byte_cnt == $cur_len)
          {
            $state = "NEW_CMD";
            if($dbg_verbose) { print CMD_F "\n----------------------------------------\n"; }

          }
          else
          {
            if($addr_11_0_value == 0xFFF)
            {

            	if($cur_len == 0x10)
            	{
                $state = "CHK_END";
              }
              else
              {
            	  print CMD_F "Error : \$addr_11_0_value=0xFFF --> cross bank boundary\n";
            	  die("Error : \$addr_11_0_value=0xFFF --> cross bank boundary\n");
            	}

            }
            $addr_11_0_value++;

          }

        }
        elsif( $state eq "CHK_END" )
        {

        }


        $bin_idx++;
      }

      $byte_cnt++;
  }

#########################################################
# DUMP REG_MAP
#########################################################

printf OUT_F<<EOF;
       00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F
---------------------------------------------------------
EOF

for($addr=$DUMP_START;$addr<($DUMP_START+$DUMP_LEN);$addr++)
{
  $addr_str = sprintf("%4.4x", $addr);

  if (($addr%16) ==0)
  {
    printf OUT_F "$addr_str : "
  }

  if($data{$addr_str} eq "")
  { printf OUT_F "XX "; }
  else
  { printf OUT_F "$data{$addr_str} "; }

  if (($addr%16) == 15)
  { printf OUT_F "\n"; }

}



  close OUT_F;
