
use Getopt::Std;
$prog_name = "ERIE ROM code parser";
$prog_rev  = "0.0.2";
$prog_cmd  = "rom_parser";

$binmode  = 1;
$override = 0;
$dbg_crc  = 0;

$rom_source = "rom_data.src";
$rom_file   = "rom_data.mem";
$rom_file   = "rom_data.bin" if $binmode;

$CRC_INIT_VALUE = "12";
$DMA_XOR_VALUE  = "00";

$hw_set_xor_value = 0;
$dma_xor_value  = hex($DMA_XOR_VALUE);

$enter_byte_num = 16; # set 1 for gen test pattern

#---------------------------------------------------
# Get options
#---------------------------------------------------
# -m    : ROM mode
# -w    : Override bin file without check target exist
# -t    : output TXT mode
# -d    : defines
# -i    : ROM source
# -k    : Debug mode
# -f    : FW file name
# -a    : FW addr
# -h    : Append header file
# -o    : Output file name
# -v    : Gen test vector
# -b    : Append 0xFF to boundary

&getopts('i:o:wtd:m:kf:a:h:vs:l:b:') || &usage();

$override         = 1 if defined($opt_w);
$binmode          = 0 if defined($opt_t);

# $ROM_MODE       = $opt_m if defined($opt_m);
$defines          = $opt_d if defined($opt_d);
$rom_source       = $opt_i if defined($opt_i);
$rom_file         = $rom_source;
$debug            = $opt_k;
$gen_test_vector  = $opt_v;

$fw_file          = $opt_f if defined($opt_f);
$fw_begin_addr    = $opt_a if defined($opt_a);
$fw_begin_addr    =~ s/0x//;
$fw_begin_addr    = hex($fw_begin_addr);
$DMA_ROM_ADDR     = sprintf("%8.8x",$fw_begin_addr);
$hdr_file         = $opt_h if defined($opt_h);

$dmp_st           = "0000";
$dmp_st           = $opt_s if defined($opt_s);
$dmp_st           =~ s/t_dmp=//;

$dmp_len          = "0";
$dmp_len          = $opt_l if defined($opt_l);
$dmp_len          =~ s/en_dmp=//;

$append_boundary  = 0;
$append_boundary  = $opt_b if defined($opt_b);

if($gen_test_vector)
{
  $binmode=0;
  $enter_byte_num=1;
  ${byte_prefix}  = "`SPI_BOOT_BFM_MEM[idx]    = 8'h";
  ${byte_postfix} = "; idx++;";
}
else
{
  ${byte_prefix}  = "";
  ${byte_postfix} = "";
}

#####################################################
# Assign output file
#####################################################

if( $binmode ) {
  $rom_file   =~ s/\.[a-zA-Z0-9_]*$/\.bin/;
} else {
  $rom_file   =~ s/\.[a-zA-Z0-9_]*$/\.mem/;
}
$rom_file   = $opt_o if defined($opt_o);
# $rom_file   =~ s/\./_$ProductID\./ if defined($opt_p);


if( -e $rom_file && $override == 0 ) {
  &err("Output file $rom_file exists!");
}
$debug_output = $rom_file.".dbg" if $debug;

print <<EOF;
===============================================

ProductID        =  $ProductID
defines          =  $defines
rom_source       =  $rom_source
rom_file         =  $rom_file

hdr_file         =  $hdr_file

fw_file          =  $fw_file
fw_begin_addr    =  $fw_begin_addr

===============================================
EOF

#---------------------------------------------------
# Read FW and cal CRC
#---------------------------------------------------

$max_fw_addr           = 0;
$total_fw_byte_cnt     = 0;
$total_fw_byte_cnt_str = "0000";
$last_fw_addr          = 0;
$cal_crc               = "00";


if($fw_file ne "")
{
  
  open(IN_FW,$fw_file) or die "Cannot open file $fw_file due to $!";

  if($fw_file =~ /\.bin$/)
  {
    binmode(IN_FW);
    $fw_idx=0;
    while(read(IN_FW,$byte,1)) 
    {
      $fw_raw[$fw_idx] = sprintf ("%2.2x",ord($byte));
      # print "$fw_raw[$fw_idx]\n";
      
      $fw_idx++;
      
      $max_fw_addr       = $fw_idx;
      $total_fw_byte_cnt = $fw_idx;
      $last_fw_addr      = $fw_begin_addr + $fw_idx - 1;
    
    }
    
    # print "==== fw_idx=$fw_idx\n";
    
  }
  elsif($fw_file =~ /\.hex$/)
  {

    @fw_lines = <IN_FW>;
  
    foreach $fw_line (@fw_lines)
    {
  
      # print "$fw_line";
  
      if($fw_line =~ /\:(\w\w)(\w\w\w\w)(\w\w)(\w+)(\w\w)/)
      {
        $fw_byte_cnt    = hex($1);
        $fw_addr        = hex($2);
        $fw_record_type = $3;
        $fw_data        = $4;
        $fw_checksum    = $5;
  
  # print <<EOF;
  #   --> $fw_line
  #
  #   fw_byte_cnt    = $fw_byte_cnt;
  #   fw_addr        = $fw_addr;
  #   fw_record_type = $fw_record_type;
  #   fw_data        = $fw_data;
  #   fw_checksum    = $fw_checksum;
  #
  # EOF
  #      <STDIN>;
  
        # Check record type must be 0
        if($fw_record_type != 0)
        { print "Error : fw_record_type($fw_record_type) != 0, fw_line=$fw_line";
          exit;
        }
  
        # Check fw_byte_cnt and $fw_data length
        if(length($fw_data)/2 != $fw_byte_cnt)
        { print "Error : fw_data($fw_data) not match with fw_byte_cnt($fw_byte_cnt)\n";
          exit;
        }
  
        # Assign $max_fw_addr
        if($fw_addr > $max_fw_addr)
        {
          $max_fw_addr          = $fw_addr;
          $total_fw_byte_cnt    = $fw_addr + $fw_byte_cnt;
          $last_fw_addr         = $fw_begin_addr + $max_fw_addr + $fw_byte_cnt - 1;
        }
  
  
        for($fw_idx=$fw_addr;$fw_idx<$fw_addr+$fw_byte_cnt;$fw_idx++)
        {
          $fw_data =~ s/^(\w\w)//;
          $byte   =$1;
  
          if($fw_raw[$fw_idx] ne "")
          { print "Error : \$fw_raw[$fw_idx]($fw_raw[$fw_idx]) not empty before write\n"; exit; }
          else
          { $fw_raw[$fw_idx] = $byte;
            # print "fw_raw[$fw_idx]=$fw_raw[$fw_idx]\n";
          }
        }
      }
    }
  } # end of elsif($fw_file =~ /\.hex$/) 
  else
  {
    die("Error : Unknown FW file extension\n");
  }
  
  #cal_crc
  ini_crc($CRC_INIT_VALUE);

  # cal LEN CRC
  $total_fw_byte_cnt_str = sprintf("%4.4x",$total_fw_byte_cnt);
  cal_crc(substr($total_fw_byte_cnt_str,2,2));
  cal_crc(substr($total_fw_byte_cnt_str,0,2));

  # cal data CRC
  for($fw_idx=0;$fw_idx<$total_fw_byte_cnt;$fw_idx++)
  { cal_crc($fw_raw[$fw_idx]); }

  $cal_crc = get_crc();

  print "total_fw_byte_cnt=$total_fw_byte_cnt\n";
  print "cal_crc=$cal_crc\n";

}


#---------------------------------------------------
# Parsing input file
#---------------------------------------------------

$section = "NOUSE";
$byte  = 0;
%version = {};
@if_match = ();
@if_write = ();
$if_count = 0;

$if_match[0] = 1;
$if_write[0] = 1;
$version{FOUND} = 0;

$count_symbol = "ghijklmnopqrstuvwxyz";
#@order = ( "PCIE_CONFIG_HIGH","USB_CONFIG_HIGH","PCIE_CONFIG_LOW","USB_CONFIG_LOW","MISC_CONFIG","END");
if( $#order == -1 ) {
  @order = ( "JUMP_OFFSET","ROM_VERSION","DEVICE_ID","BEFORE_MAIN_RESET","MISC_CONFIG","MISC_NEED_PECLK","END");
}

$section = "NOUSE";
open(IN,$rom_source) or die "Cannot open file $rom_source due to $!";
if( $debug ) {
  if( !open(DBG,">$debug_output") ) {
    die "Cannot open file $debug_output due to $!";
    $debug = 0;
  }
}
while( <IN> ) {
  chomp($_);
  %version = &check_version($_) unless $version{FOUND};
  $_ =~ s/#.*$//;
  $_ =~ s/[ \t]//g;
  next if $_ =~ /^$/;
  ######################### Conditioning ##############################
  # process &if,&else,&elsif,&endif
  if( $_ =~ /\&if([\w\|]+);/ ) {
    $condition = $1;
    @cond = split(/\|/,$condition);
    $if_count ++;
            print DBG ">"x$if_count."$condition" if $debug; ## ----------------------------------
    foreach $condition ( @cond ) {
      if( $defines =~ /\b$condition\b/ ) {
        $if_match[$if_count] = $if_match[$if_count-1];
        $if_write[$if_count] = $if_match[$if_count-1];
            print DBG " <- matched" if $debug; ## ----------------------------------
        last;
      } else {
        $if_match[$if_count] = 0;
        $if_write[$if_count] = 0;
      }
    }
            print DBG "\n" if $debug; ## ----------------------------------
    next;
  } elsif( $_ =~ /\&else;/ ) {
    if( $if_write[$if_count] == 0 ) {
      $if_match[$if_count] = $if_match[$if_count-1];
      $if_write[$if_count] = $if_match[$if_count-1];
            print DBG "-"x$if_count."\n" if $debug; ## ----------------------------------
    } else {
      $if_match[$if_count] = 0;
    }
    next;
  } elsif( $_ =~ /\&elsif([\w\|]+);/ ) {
            print DBG "-"x$if_count."$condition" if $debug; ## ----------------------------------
    $condition = $1;
    if( $if_write[$if_count] == 0 ) {
      @cond = split(/\|/,$condition);
      foreach $condition ( @cond ) {
        if( $defines =~ /\b$condition\b/ ) {
          $if_match[$if_count] = $if_match[$if_count-1];
          $if_write[$if_count] = $if_match[$if_count-1];
          last;
        }
      }
    } else {
      $if_match[$if_count] = 0;
    }
    next;
  } elsif( $_ =~ /\&endif;/ ) {
    $if_match[$if_count] = 0;
            print DBG "<"x$if_count."\n" if $debug; ## ----------------------------------
    $if_count --;
    next;
  }
  #######################################################################
  if(  $if_match[$if_count] )  {

    # Proc &xxx (except &if,&else,&elsif,&endif)
    if( $_ =~ /\&(\w+);/ ) {
      $keyword = $1;

      # Proc &DWORD_BOUNDARY
      if( $keyword eq "DWORD_BOUNDARY" ) {
        $byte = length($ROM{$section});
        $byte = $byte % 32 if $byte > 0;
        $ROM{$section} = "0"x(32-$byte).$ROM{$section} if $byte > 0;
      
      # Proc &BYTE_BOUNDARY
      } elsif( $keyword eq "BYTE_BOUNDARY" ) {
        $byte = length($ROM{$section});
        $byte = $byte % 8 if $byte > 0;
        print "==== "."0"x(8-$byte)." ====";
        $ROM{$section} = "0"x(8-$byte).$ROM{$section} if $byte > 0;
      
      # Proc &INSERt_BIN_FILE [TimLin]
      } elsif( $keyword eq "INSERt_BIN_FILE" ) {
        if($_ =~ /\&INSERt_BIN_FILE;\$file=(\S+)\$target=(\w+)\$skip_byte_num=(\w+)/)
        { 
        	$bin_file_name     = $1;
        	$target_addr_str   = $2;
        	$skip_byte_num_str = $3;
        	
        	insert_bin_file($bin_file_name, $target_addr_str, $skip_byte_num_str); 
        }
        else
        {
          print "Error : INSERt_BIN_FILE format error\n";
          print "        format --> &INSERt_BIN_FILE; file=file_name target=target_addr\n";
          print "        input  --> $_\n";
          die;
        }
      # Get section name
      } else {
        $section = $keyword;
        $ROM{$section} = (0,"");
        $dword = 0;
        $byte  = 0;
      }

    # Get field into
    } elsif($_ =~ /^(\w+)=(\d+)\((\w+)\)/) {
      $field = $1;
      $digit = $2;
      $value = $3;
      $count_symbol =~ s/(\w)$//;
      $symbol = "$1"x$digit;
      $ROM{$section} = $symbol.$ROM{$section};
      $COUNT{$field}{VALUE}  = $value;
      $COUNT{$field}{SYMBOL} = $symbol;
      $COUNT{$field}{DIGIT}  = $digit;

    # Replace CHECK keyword value from input file
    } else {
      if($_ =~ /^CompID=CHECK\b/        ) { $_ =~ s/\bCHECK\b/8\'h$version{CompID}/;    }
      if($_ =~ /^ProductID=CHECK\b/     ) { $_ =~ s/\bCHECK\b/8\'h$version{ProductID}/; }
      if($_ =~ /^FormalVER=CHECK\b/     ) { $_ =~ s/\bCHECK\b/8\'h$version{FormalVER}/; }
      if($_ =~ /^SubVER=CHECK\b/        ) { $_ =~ s/\bCHECK\b/8\'h$version{SubVER}/;    }

      # check CRC
      if($_ =~ /^CRC_INIT_VALUE=CHECK\b/) { $_ =~ s/\bCHECK\b/8\'h${CRC_INIT_VALUE}/;
                                           print "$_\n";
                                          }
      if($_ =~ /^DMA_ROM_ADDR=CHECK\b/  ) { $_ =~ s/\bCHECK\b/32\'h${DMA_ROM_ADDR}/;
                                             print "$_\n";
                                          }
      if($_ =~ /^DMA_LEN=CHECK\b/       ) { $_ =~ s/\bCHECK\b/16\'h${total_fw_byte_cnt_str}/;
                                             print "$_\n";
                                          }
      if($_ =~ /^CHK_CRC_VALUE=CHECK\b/ ) { $_ =~ s/\bCHECK\b/8\'h${cal_crc}/;
                                             print "$_\n";
                                          }
      if($_ =~ /^DMA_XOR_VALUE=CHECK\b/ ) { $_ =~ s/\bCHECK\b/8\'h${DMA_XOR_VALUE}/;
                                             print "$_\n";
                                          }

      $_ =~ /^(\w+)=(\d+)\'([hbd])([0-9a-fA-F_]+)/;
      $field = $1;
      $digit = $2;
      $radix = $3;
      $value = $4;
      $value =~ s/_//g;
      # print "$_ => $field : $digit : $radix : $value\n";
            print DBG "## $_ => $field : $digit : $radix : $value\n" if $debug; ## ----------------------------------
         if( $radix eq "d" ) {
          print "\n!![WARNING] Problem found in $field that it should be in decimal ($value).\n" unless $value =~ /^[0-9]+$/;
          $value = Dec2Num($value,2);
      } elsif( $radix eq "h" ) {
          $value = Hex2Num($value,2);
      } else                   {
          print "\n!![WARNING] Problem found in $field that it should be in binary ($value).\n" unless $value =~ /^[01]+$/;
      }
      if( $field eq "ProductID" && defined($ProductID)) {
        $value = Hex2Num($ProductID,2);
      }
      
      print DBG "<-- $_  => $value\n";
      
      $value = LeadingSymbol("0",$digit,$value);
      print DBG "--> $_  => $value\n";
      $ROM{$section} = $value.$ROM{$section};
    }
  }
}

close(IN) or &err("ERROR: Cannot close file $rom_source due to $!");

if( !defined($ROM{CHECK_VALUE}))
{
  print "CHECK_VALUE has to be defined!\n";
  <STDIN>;
  exit;
}

$fn = length($ROM{CHECK_VALUE});
for($i=0;$i<@order;$i++) {
  foreach $field ( keys(%COUNT) ) {
    if( $order[$i] eq $COUNT{$field}{VALUE} ) {
      $COUNT{$field}{VALUE} = LeadingSymbol("0",$COUNT{$field}{DIGIT},Num2Num(10,int($fn/8),2));
    }
  }
  $fn += length($ROM{$order[$i]});
}
print "\n!![WARNING] Problem found critical byte number is not a integer : ".($fn/8)."\n" if ($fn%8)>0;
#$result = "NN".$ROM{CHECK_VALUE};
$result = $ROM{CHECK_VALUE};

foreach $cfg ( @order ) {
  $result = $ROM{$cfg}.$result;
}

foreach $field ( keys(%COUNT) ) {
  if( $COUNT{$field}{VALUE} =~ /^[01]+$/ ) {
    $symbol = $COUNT{$field}{SYMBOL};
    $value  = $COUNT{$field}{VALUE};
    $result =~ s/$symbol/$value/;
  } else {
    print "\n!![WARNING] Cannot find the Section $COUNT{$field}{VALUE} when Field $field is requesting\n";
  }
}

#---------------------------------------------------
# Output to ROM file
#---------------------------------------------------

# Add HW version
$hw_fw_version = "$version{CompID}.$version{ProductID}.$version{FormalVER}.$version{SubVER}";

# Add FW version
if($fw_file ne "")
{
  $fw_version =  $fw_file;
  $fw_version =~ s/(\S+)_//;
  $fw_version =~ s/\.?(\w+)$//;
  $fw_version = "_FW_".$fw_version;
  
  $hw_fw_version .= "$fw_version";
}
else
{ $fw_version = ""; }

$rom_file   =~ s/\./_${hw_fw_version}\./;

$rom_file = uc($rom_file);
$rom_file =~ s/\.BIN/\.bin/;
$rom_file =~ s/\.MEM/\.mem/;

open(OUT,">$rom_file") or &err("Cannot open file $rom_file due to $!");
binmode OUT if $binmode;
$ff_cnt           = 6;
$fw_idx           = 0;
$set_romfetch_end = 0;
$chk_fw_overlap   = 0;
$hdr_byte_cnt     = 0;

#---------------------------------------------------
# Append hdr file to ROM file
#---------------------------------------------------

if($hdr_file ne "")
{

  # Read Header file
  open(HDR_F,$hdr_file) or die "Cannot open file $fw_file due to $!";

  @hdr_lines = <HDR_F>;
  $hdr_setting = "";

  foreach $hdr_line (@hdr_lines)
  {
    $hdr_line =~ s/[\r\n ]//g;
    $hdr_setting .= $hdr_line;
  }

  print "hdr_setting=$hdr_setting\n";

  if(length($hdr_setting)%2 == 1)
  {
    print "Error : length(\$hdr_setting)%2 == 1\n";
    die();
  }
  
  $hdr_setting_len = length($hdr_setting)/2;

  while ($hdr_setting ne "")
  {
    $hdr_setting =~ s/^(\w\w)//;
    $hdr_byte = $1;
    if( $binmode ) { print OUT pack("H2",$hdr_byte); }
    else           { print OUT "${byte_prefix}${hdr_byte}${byte_postfix}";
                     if($hdr_byte_cnt % $enter_byte_num==$enter_byte_num-1) { print OUT "\n"; }
                   }

    # print "$hdr_byte\n";
    $hdr_byte_cnt++;
  }

}



print "last_fw_addr=$last_fw_addr\n";

#$byte = LeadingSymbol("0",8,Dec2Num((($fn/8)+1),2));
#$result =~ s/NN/$byte/;
for($i=0;$i<=32768;$i++) {
  #rom fetch data
  if( $result =~ /(\w\w\w\w\w\w\w\w)$/ ) {
    $byte = $1;
    $result =~ s/$byte$//;
    $byte = LeadingSymbol("0",2,Num2Num(2,$byte,16));
    
    $hw_set_xor_value = $hw_set_xor_value ^ hex($byte);
    printf DBG " -- HW_SET_XOR[$i] = %2.2X BYTE[$i]=%2.2X\n", $hw_set_xor_value, hex($byte);
    
    if( $binmode ) { print OUT pack("H2",$byte); }
    else           { print OUT "${byte_prefix}${byte}${byte_postfix}";
                     if(($i+$hdr_byte_cnt) % $enter_byte_num==$enter_byte_num-1) { print OUT "\n"; }
                   }
  }
  # set HW_SET_XOR_VALUE
  elsif($set_set_xor_value==0)
  {
    
    $hw_set_xor_value = $hw_set_xor_value ^ hex("FC");
    printf DBG " -- HW_SET_XOR[$i] = %2.2X BYTE[$i]=%2.2X\n", $hw_set_xor_value, 0xFC;
    $hw_set_xor_value = $hw_set_xor_value ^ hex("B0");
    printf DBG " -- HW_SET_XOR[$i] = %2.2X BYTE[$i]=%2.2X\n", $hw_set_xor_value, 0xB0;
    $hw_set_xor_value = $hw_set_xor_value ^ hex("11");
    printf DBG " -- HW_SET_XOR[$i] = %2.2X BYTE[$i]=%2.2X\n", $hw_set_xor_value, 0x11;
    
    $hw_set_xor_value_str = sprintf ("%2.2X", $hw_set_xor_value);
    @HW_SET_XOR_REG = ("FC","B0","11",$hw_set_xor_value_str);
    printf DBG " -- HW_SET_XOR_VALUE = $hw_set_xor_value_str\n";
    
    foreach $byte ( @HW_SET_XOR_REG ) {
      if( $binmode ) { print OUT pack("H2",$byte); }
      else           { print OUT "${byte_prefix}${byte}${byte_postfix}";
                       if(($i+$hdr_byte_cnt) % $enter_byte_num==$enter_byte_num-1) { print OUT "\n"; }
                     }
        
      $i++;
    }
    $i--;
    $set_set_xor_value=1;
  }
  # output FFFFFFFF to stop ROM FETCH
  elsif($set_romfetch_end==0)
  {
    
    if($append_boundary != 0)
    {
      if((($i+$ff_cnt) % $append_boundary) == 0)
      {
         $append_ff_cnt = $ff_cnt;
      }
      else
      {
         $target_addr = (int(($i+$hdr_byte_cnt+$ff_cnt)/$append_boundary) + 1) * $append_boundary;
         
         $append_ff_cnt =  $target_addr - $i - $hdr_byte_cnt - $ff_cnt; # in ending will append $ff_cnt 0xFF
         
         print "hdr_byte_cnt=$hdr_byte_cnt\n";
         print "append_boundary=$append_boundary, i=$i, target_addr=$target_addr, append_ff_cnt=$append_ff_cnt\n";
         
      }
      
    }
    else
    {
      $append_ff_cnt = $ff_cnt;
    }  
    
    foreach( 1 .. $append_ff_cnt ) {
      if( $binmode ) { print OUT pack("H2","ff"); }
      else           { print OUT "${byte_prefix}ff${byte_postfix}";
                       if(($i+$hdr_byte_cnt) % $enter_byte_num==$enter_byte_num-1) { print OUT "\n"; }
                     }
    
      $i++;
    }
    
    $i--;
    $set_romfetch_end=1;
  }
  #fw_data
  elsif(($last_fw_addr != 0) && ($i<=$last_fw_addr)) {
    if($chk_fw_overlap==0)
    {
      if($i>=$fw_begin_addr)
      { print "Error : ROM FETCH data(idx=$i) overrlap firmware begin addr(addr=$fw_begin_addr)\n";
        exit;
      }
      $chk_fw_overlap=1;
    }

    if($i<$fw_begin_addr)
    {
      if( $binmode ) { print OUT pack("H2","FF"); }
      else           { print OUT "${byte_prefix}ff${byte_postfix}";
                       if(($i+$hdr_byte_cnt) % $enter_byte_num==$enter_byte_num-1) { print OUT "\n"; }
                     }
    }
    else
    {
      $byte = $fw_raw[$fw_idx];
      if($byte eq "")
      { $byte = "00"; }
      
      # Cal XOR value
      $byte_value = hex($byte);
      $xor_byte = sprintf("%2.2X",$dma_xor_value ^ $byte_value);
      # print "$byte -- $xor_byte\n";
      $byte = $xor_byte;
      
      $fw_idx++;
      if( $binmode ) { print OUT pack("H2",$byte);}
      else           { print OUT "${byte_prefix}${byte}${byte_postfix}";
                       if(($i+$hdr_byte_cnt) % $enter_byte_num==$enter_byte_num-1) { print OUT "\n"; }
                     }
    }

  }
  # ending
  else {
    foreach( 1 .. $ff_cnt ) {
      if( $binmode ) { print OUT pack("H2","ff"); }
      else           { print OUT "${byte_prefix}ff${byte_postfix}";
                       if(($i+$hdr_byte_cnt) % $enter_byte_num==$enter_byte_num-1) { print OUT "\n"; }
                       $i++;
                     }
    }
    last;
  }
}

close(OUT) or &err("Cannot close file $rom_file due to $!");


print "\nROM file $rom_file has generated !\n";

## output register map

if($dmp_len ne "0")
{

$rom_file_name = $rom_file;
$rom_file_name =~ s/\.bin$//;

$REG_DUMP_CMD  = "tinyperl20\\tinyperl.exe .\\Program\\titan_bin_to_reg_map.pl ";
$REG_DUMP_CMD .= "-in_file=${rom_file_name}.bin ";
$REG_DUMP_CMD .= "-skip_byte=${hdr_setting_len} ";
$REG_DUMP_CMD .= "-reg_map_file=${rom_file_name}.reg_map ";
$REG_DUMP_CMD .= "-reg_cmd_file=${rom_file_name}.reg_cmd ";
$REG_DUMP_CMD .= "-dmp_st=${dmp_st} -dmp_len=${dmp_len}";

system("$REG_DUMP_CMD");

}

# if( !defined($id) ) { &pause; }

##################################################################################################
##################################################################################################
##################################################################################################
##################################################################################################
##################################################################################################

sub usage {
  unless( -t STDOUT) {
    select(STDERR);
  }
  print "\n";
  print "$prog_name rev$prog_rev\n";
  print "Usage: $prog_cmd -i Filename [-o Filename] [-k]\n";
  print "  -i : Select Input Filename\n";
  print "  -d : Defines, Seperate by |\n";
  print "  -p : Product ID (1 to f)\n";
  print "  -o : Specify Output Filename if you want.\n";
  print "       Default output filename will be the input filename\n";
  print "       with extension replaced by '.bin'.\n";
  print "  -w : Override output file if it exists.!\n";
  print "  -t : Generate Text file!\n";
  # print "  -m : ROM_MODE\n";
  print "  -f : FW file name\n";
  print "  -a : FW addr\n";
  print "  -h : Append header file\n";
  print "\n\n";

  exit 1;
}

sub check_version {  ## String
  my $string = $_[0];
  my %version = {};
  $string =~ s/\t/ /g;
  $string =~ s/\s+/ /g;
  $string =~ s/^\s+//g;
  $string =~ s/\s+$//g;
  if( $string =~ /^#+ ([0-9a-fA-F][0-9a-fA-F])\.([0-9a-fA-F][0-9a-fA-F])\.([0-9a-fA-F][0-9a-fA-F])\.([0-9a-fA-F][0-9a-fA-F])/ ) {
    $version{FOUND}      = 1;
    $version{CompID}     = $1;
    $version{ProductID}  = $2;
    $version{FormalVER}  = $3;
    $version{SubVER}     = $4;
    
    print "version{ProductID} = $version{ProductID}\n";
    
    # $version{ProductID} = "0" if $version{ProductID} =~ /x/;
  }
  
  # $version{ROM_MODE} = $ROM_MODE;
  
  return %version;
}


##################################################################################################################
sub set {
  return &SetField(@_);
}
sub SetField { ## Hex, Range, value
  my @bin = reverse(split(//,&LeadingSymbol("0",32,&Num2Num(16,$_[0],2))));
  my $range = $_[1];
  my @value = reverse(split(//,&Num2Num(16,$_[2],2)));
  my $out = "";
  my ($beg,$end) = (0,0);
  my $idx = 0;
  if( $range =~ /(\d+):(\d+)/ ) {
    if( $1 < $2 ) {
      $beg = $1;
      $end = $2;
    } else {
      $beg = $2;
      $end = $1;
    }
    foreach( $beg .. $end ) {
      $idx = $_ - $beg;
      if( defined $value[$idx] ) {
        $bin[$_] = $value[$idx];
      } else {
        $bin[$_] = 0;
      }
    }
  } else {
    $bin[$range] = $value[0];
  }
  $_[0] = &Num2Num(2,join("",reverse(@bin)),16);
  return $_[0];
}

sub get {
  return &GetField(@_);
}
sub GetField { ## Dec Range
  my @bin = reverse(split(//,&Num2Num(16,$_[0],2)));
  my $msb = $#bin;
  my $range = $_[1];
  my $out = "";
  my ($beg,$end) = (0,0);
  if( $range =~ /(\d+):(\d+)/ ) {
    if( $1 < $2 ) {
      $beg = $1;
      $end = $2;
    } else {
      $beg = $2;
      $end = $1;
    }
    foreach( $beg .. $end ) {
      last if( $_ > $msb );
      $out = $bin[$_].$out;
    }
  } else {
    $out = $bin[$range] unless $range > $msb;
  }
  return &Num2Num(2,$out,16);
}


sub keyin {
  print $_[0]." > ";
  my $in = <STDIN>;
  chomp($in);
  $in =~ s/^[ \t]+//;
  $in =~ s/[ \t]+$//;
  return $in;
}

sub warn {
  print "[*]WARNING: ".$_[0];
# print "\n";
  <STDIN>;
}

sub err {
  print "[*]ERROR  : ".$_[0]."\n";
  &end unless $_[1] == -1;
}

sub end {
  pause(".... Thank you !");
  exit;
}

sub pause { # msg
  print @_;
  <STDIN>;
}

##################################################################################################################
sub h2d { # ( IN )
  return &Num2Dec(16,$_[0]);
}
sub d2h { # ( IN )
  return &Dec2Num($_[0],16);
}
sub h2b { # ( IN )
  return &Num2Num(16,$_[0],2);
}
sub b2h { # ( IN )
  return &Num2Num(2,$_[0],16);
}
sub d2b { # ( IN )
  return &Dec2Num($_[0],2);
}
sub b2d { # ( IN )
  return &Num2Dec(2,$_[0]);
}
sub fxdw {
  return "0x".&FormDWord(@_);
}
sub fdw {
  return &FormDWord(@_);
}
sub fxb {
  return "0x".&FormByte(@_);
}
sub fb{
  return &FormByte(@_);
}


sub FormDWord { # IN, Radix
  my $hex = $_[0];
  return &LeadingSymbol("0",8,$hex);
}
sub FormByte { # IN, Radix
  my $hex = $_[0];
  return &LeadingSymbol("0",2,$hex);
}

sub Hex2Num { # ( Number, Output Radix )
  return &Num2Num(16,$_[0],$_[1]);
}

sub Hex2Dec { # ( IN )
  return &Num2Dec(16,$_[0]);
}

sub Dec2Hex { # ( IN )
  return &Dec2Num($_[0],16);
}

sub Num2Num { # ( Input_Radix, Number, Output Radix )
  my $dec = &Num2Dec($_[0],$_[1]);
  return &Dec2Num($dec,$_[2]);
}


sub Num2Dec { # ( Input_Radix , Number)
  my $radix = $_[0];
  my $num = lc($_[1]);
     $num = $1 if $num =~ m/0x([0-9a-f]+)/;
  my @digit = split(//,$num);
  my $out = 0;
  foreach $d ( @digit ) {
       if( $d eq "0" ) { $out = $out * $radix +  0; }
    elsif( $d eq "1" ) { $out = $out * $radix +  1; }
    elsif( $d eq "2" ) { $out = $out * $radix +  2; }
    elsif( $d eq "3" ) { $out = $out * $radix +  3; }
    elsif( $d eq "4" ) { $out = $out * $radix +  4; }
    elsif( $d eq "5" ) { $out = $out * $radix +  5; }
    elsif( $d eq "6" ) { $out = $out * $radix +  6; }
    elsif( $d eq "7" ) { $out = $out * $radix +  7; }
    elsif( $d eq "8" ) { $out = $out * $radix +  8; }
    elsif( $d eq "9" ) { $out = $out * $radix +  9; }
    elsif( $d eq "a" ) { $out = $out * $radix + 10; }
    elsif( $d eq "b" ) { $out = $out * $radix + 11; }
    elsif( $d eq "c" ) { $out = $out * $radix + 12; }
    elsif( $d eq "d" ) { $out = $out * $radix + 13; }
    elsif( $d eq "e" ) { $out = $out * $radix + 14; }
    elsif( $d eq "f" ) { $out = $out * $radix + 15; }
    else               { $out = $out * $radix;      }
  }
  return $out;
}

sub Dec2Num { # ( Dec_Number, Target_Radix )
  my $d     = $_[0];
  my $radix = $_[1];
  my $r = 0;
  my $out = "";
  while( $d > 0 ) {
    $r = $d % $radix;
    $d = ($d - $r) / $radix;
       if( $r ==  0 ) { $out = "0".$out; }
    elsif( $r ==  1 ) { $out = "1".$out; }
    elsif( $r ==  2 ) { $out = "2".$out; }
    elsif( $r ==  3 ) { $out = "3".$out; }
    elsif( $r ==  4 ) { $out = "4".$out; }
    elsif( $r ==  5 ) { $out = "5".$out; }
    elsif( $r ==  6 ) { $out = "6".$out; }
    elsif( $r ==  7 ) { $out = "7".$out; }
    elsif( $r ==  8 ) { $out = "8".$out; }
    elsif( $r ==  9 ) { $out = "9".$out; }
    elsif( $r == 10 ) { $out = "a".$out; }
    elsif( $r == 11 ) { $out = "b".$out; }
    elsif( $r == 12 ) { $out = "c".$out; }
    elsif( $r == 13 ) { $out = "d".$out; }
    elsif( $r == 14 ) { $out = "e".$out; }
    elsif( $r == 15 ) { $out = "f".$out; }
    else              { $out = "x".$out; }
  }
  $out = "0" if $out eq "";
  return $out;
}

# ------------------------------------------------

sub LeadingSymbol { # ( Symbol,  Number_Of_Digit, IN )
  my $sym = @_[0];
  my $dgt = @_[1];
  my $in  = @_[2];

  $dgt = $dgt - length($in);
  return $sym x $dgt .$in;
}

# ------------------------------------------------
sub h2b_8bit
{
  my($in_hex_byte) = @_;

  my($in_byte,$idx,@binary_ary);

  $in_byte = hex($in_hex_byte);

  for($idx=0;$idx<8;$idx++)
  { $binary_ary[$idx]= ($in_byte>>$idx) & 0x1; }

  return @binary_ary;

}

# ------------------------------------------------
sub ini_crc
{
  my($ini_crc) = @_;

  my(@ini_crc_ary,$idx);

  @ini_crc_ary = h2b_8bit($ini_crc);
  for($idx=0;$idx<8;$idx++)
  { $cur_crc[$idx] = $ini_crc_ary[$idx]; }

  $crc_idx=0;

  if($dbg_crc) {print "cur_crc = @cur_crc\n"; }

}

# ------------------------------------------------

sub cal_crc
{

  my ($in_data) = @_;

  @rom_d = h2b_8bit($in_data);

  # print "in_data=$in_data, rom_d = @rom_d\n";

#  Purpose : synthesizable CRC function
#    * polynomial: (0 4 5 8)
#    * data width: 8
#
#  Info : tools@easics.be
#         http://www.easics.com
# //////////////////////////////////////////////////////////////////////////////

  $next_crc[0] =   ((((((($rom_d[6] xor $rom_d[4]) xor $rom_d[3]) xor  $rom_d[0]) xor $cur_crc[0]) xor $cur_crc[3]) xor $cur_crc[4]) xor $cur_crc[6]);

  # print "=== $next_crc[0] = $rom_d[6] xor $rom_d[4] xor $rom_d[3] xor  $rom_d[0] xor $cur_crc[0] xor $cur_crc[3] xor $cur_crc[4] xor $cur_crc[6];\n";

  $next_crc[1] =   ((((((($rom_d[7] xor $rom_d[5]) xor $rom_d[4]) xor  $rom_d[1]  ) xor $cur_crc[1]) xor $cur_crc[4]) xor $cur_crc[5]) xor $cur_crc[7]);
  $next_crc[2] =     ((((($rom_d[6] xor $rom_d[5]) xor $rom_d[2]) xor  $cur_crc[2]) xor $cur_crc[5]) xor $cur_crc[6]);
  $next_crc[3] =     ((((($rom_d[7] xor $rom_d[6]) xor $rom_d[3]) xor  $cur_crc[3]) xor $cur_crc[6]) xor $cur_crc[7]);
  $next_crc[4] =   ((((((($rom_d[7] xor $rom_d[6]) xor $rom_d[3]) xor  $rom_d[0]  ) xor $cur_crc[0]) xor $cur_crc[3]) ,xor $cur_crc[6]) xor $cur_crc[7]);
  $next_crc[5] = ((((((((($rom_d[7] xor $rom_d[6]) xor $rom_d[3]) xor  $rom_d[1]  ) xor $rom_d[0]  ) xor $cur_crc[0]) ,xor $cur_crc[1]) xor $cur_crc[3]) xor $cur_crc[6]) xor $cur_crc[7]);
  $next_crc[6] =   ((((((($rom_d[7] xor $rom_d[4]) xor $rom_d[2]) xor  $rom_d[1]  ) xor $cur_crc[1]) xor $cur_crc[2]) ,xor $cur_crc[4]) xor $cur_crc[7]);
  $next_crc[7] =     ((((($rom_d[5] xor $rom_d[3]) xor $rom_d[2]) xor  $cur_crc[2]) xor $cur_crc[3]) xor $cur_crc[5]);

  @cur_crc = @next_crc;

  # printf "cur_crc = %x\n",get_crc();

  if($dbg_crc) { print "idx=$crc_idx, in_data=$in_data, "; }
  $crc_idx++;
  get_crc();

  # exit();
}

# ------------------------------------------------

sub get_crc
{

  my($out_crc_bin,$out_crc, $out_crc_hex,$idx);

  $out_crc=0;
  for($idx=0;$idx<8;$idx++)
  {
    if($cur_crc[$idx]==1)
    { $out_crc += 1<<$idx; }

    # print "cur_crc[$idx]=$cur_crc[$idx], out_crc = $out_crc\n";

  }

  $out_crc_hex = sprintf("%2.2x",$out_crc);

  if($dbg_crc) { printf "out_crc_hex=%s\n",$out_crc_hex; }

  return $out_crc_hex;

}

sub insert_bin_file
{

  my ($in_bin_file, $in_addr, $skip_byte_num_str) = @_; 
  my (@bin_raw, $bin_idx, $byte, $skip_byte_num);
  
  print "\ninsert_bin_file($in_bin_file, $in_addr, $skip_byte_num_str)\n";
  print DBG "\ninsert_bin_file($in_bin_file, $in_addr, $skip_byte_num_str)\n";
  
  open(IN_BIN,$in_bin_file) or die "Cannot open file $in_bin_file due to $!";
  
  binmode(IN_BIN);
  
  if($skip_byte_num_str =~ s/0x//)
  { $skip_byte_num = hex($skip_byte_num_str); }
  else
  { $skip_byte_num = $skip_byte_num_str; }
  
  
  $bin_idx=0;
  $byte_cnt=0;
  while(read(IN_BIN,$byte,1)) 
  {
      
      if($byte_cnt>=$skip_byte_num) 
      {
        $bin_raw[$bin_idx] = sprintf ("%2.2x",ord($byte));
        print DBG "$bin_raw[$bin_idx] ";
        if($bin_idx%16==15)
        { print DBG "\n"; }
        
        $bin_idx++;
      }
      
      $byte_cnt++;
  }
  
  print DBG "\n";
  
  if($in_addr =~ s/0x//)
  { $in_addr_str = sprintf ("%4.4X", hex($in_addr)); }
  else
  { $in_addr_str = sprintf ("%4.4X", $in_addr); }
  
  $bank_str   = substr ($in_addr,0,1);
  $offset     = hex(substr($in_addr,1,3));
  
  printf DBG ("\n----- bank_str=%s offset=%x ----\n\n", $bank_str, $offset);
  
  # BANK
  
  $value = Hex2Num("$bank_str"."C",2);
  $value = LeadingSymbol("0",8,$value);
  
  $ROM{$section} =  $value.$ROM{$section};
  
  $remain_bytes = $bin_idx;
  
  $bin_idx=0;
  while ($remain_bytes>0) 
  {
    
    $offset_0   = sprintf("%1.1X", ($offset>>8)  & 0xF);
    $offset_1   = sprintf("%1.1X", ($offset>>4)  & 0xF);
    $offset_2   = sprintf("%1.1X", ($offset>>0)  & 0xF);
    
    printf DBG "%3.3X --> $offset_0 -- $offset_1 -- $offset_2\n", $offset;
    
    if($remain_bytes>=16)
    {
      add_ROM_by_2Digit_str($offset_2."F");
      add_ROM_by_2Digit_str($offset_0.$offset_1);
      
      for($byte_cnt=0;$byte_cnt<16;$byte_cnt++)
      {
        add_ROM_by_2Digit_str($bin_raw[$bin_idx]);
        print DBG "ADDR_${offset} = $bin_raw[$bin_idx]\n";
        $bin_idx++;
        $offset++;
      }
      $remain_bytes -= 16;
    }
    elsif($remain_bytes>=8)
    {
      add_ROM_by_2Digit_str($offset_2."7");
      add_ROM_by_2Digit_str($offset_0.$offset_1);
      
      for($byte_cnt=0;$byte_cnt<8;$byte_cnt++)
      {
        add_ROM_by_2Digit_str($bin_raw[$bin_idx]);
        print DBG "ADDR_${offset} = $bin_raw[$bin_idx]\n";
        $bin_idx++;
        $offset++;
      }
      $remain_bytes -= 8;
    }
    elsif($remain_bytes>=4)
    {
      add_ROM_by_2Digit_str($offset_2."3");
      add_ROM_by_2Digit_str($offset_0.$offset_1);
      
      for($byte_cnt=0;$byte_cnt<4;$byte_cnt++)
      {
        add_ROM_by_2Digit_str($bin_raw[$bin_idx]);
        print DBG "ADDR_${offset} = $bin_raw[$bin_idx]\n";
        $bin_idx++;
        $offset++;
      }
      $remain_bytes -= 4;
    }
    else
    {
      add_ROM_by_2Digit_str($offset_2."0");
      add_ROM_by_2Digit_str($offset_0.$offset_1);
      
      add_ROM_by_2Digit_str($bin_raw[$bin_idx]);
      print DBG "ADDR_${offset} = $bin_raw[$bin_idx]\n";
      $bin_idx++;
      $offset++;
      $remain_bytes--;
    
    }
  }
}

sub add_ROM_by_2Digit_str
{
  my($in_str) = @_;
  my($value);
  
  if(length($in_str) != 2)
  {
    print "Error : add_ROM_by_2Digit_str($in_str) --> length of in_str is not 2\n";
    die;
  }
     
  $value = Hex2Num($in_str,2);
  $value = LeadingSymbol("0",8,$value);
  
  $ROM{$section} =  $value.$ROM{$section};
  	
}


1;

